/*
Google Music Frame <http://Projects.Fenryxo.cz/Google_Music_Frame/>

Copyright 2011 Jiří Janoušek <janousek.jiri@gmail.com>

This program is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License version 3, as published 
by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranties of 
MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using Gtk;
using Fenryxo.Logger;
using Fenryxo.Config;
using GMF.Helpers;

namespace GMF{
	
	/**
	 * The main application class stores metadata (application name, etc.),
	 * configuration (including UI state) and
	 * launches other modules and handles basic actions like activate/quit
	 */
	public class Application: Fenryxo.Application {		
		private	MainWindow main_window;
		private Frame frame;
		private Player player;
		private Storage storage;
		private GMF.MediaKeys.Manager media_keys;
		TrayIcon tray_icon;	
		private bool hide_when_playing = true;
		
		[Description(nick = "Configuration", blurb = "Dictionary storing all Google Music Frame configuration.")]
		public ConfigDict config{ get; private set; }		
		
		[Description(nick = "UI State", blurb = "Dictionary storing state of user interface (window dimensions, etc.)")]
		public ConfigDict ui_state{ get; private set; }	
		

#if GLIB_DBUS
		private DockManager dock_manager;
		private GMF.MPRIS.Manager mpris_manager;	
#endif	
#if UNITY_QUICK_LIST
		GMF.Helpers.UnityLauncher unity_quick_list;
#endif
		
		[Description(nick = "Whether hide on close is possible", blurb = "If true, the main application application window is hidden instead of being destroyed.")]
		public bool can_hide{
			get{
				log_debug(@"Config: hide_when_playing: $hide_when_playing");
				return this.hide_when_playing
				&& this.player.playback_state == "playing"
				&& this.tray_icon != null
				&& this.tray_icon.can_hide; 
			}
		}
		
		
		/**
		 * Constructor for the main application class
		 * 
		 * @param config Configuration
		 * @param ui_state User interface state
		 * @param storage File storage
		 */
		public Application(ConfigDict config, ConfigDict ui_state, Storage storage){
			base(GMF.Config.UNIQUE_NAME, GMF.Config.DESKTOP_ENTRY, GMF.Config.APPNAME, null);
			this.icon = GMF.Config.APP_ICON;
			this.version = GMF.Config.VERSION;
			this.app_name = GMF.Config.APPNAME;
			this.display_name = GMF.Config.DISPLAY_NAME;
			this.config = config;
			this.ui_state = ui_state;
			this.storage = storage;
			this.hide_when_playing = config.bool("hide_when_playing", true);
			this.config.notify_changed.connect(this.config_changed);
			this.init_ui();
			
			// Player provides playback controls			
			this.player = new GMF.Player(this, this.frame, this.storage);
			
#if GLIB_DBUS
			this.mpris_manager = new GMF.MPRIS.Manager(this, this.player);
			this.dock_manager = new DockManager(this, this.player);
#endif		
			this.tray_icon = new TrayIcon(this, this.player);
		}
		
		~ Application(){
			log_debug("~Application");
		}
		
		/**
		 * Initialize user interface
		 */
		private void init_ui(){
			this.frame = new Frame(this);
			this.main_window = new MainWindow(this);
			this.main_window.close_action.connect(this.on_close_action);
			this.main_window.go_home_action.connect(this.frame.go_home);
			this.main_window.refresh_action.connect(this.frame.reload);
			this.main_window.settings_action.connect(() => {
				SettingsWindow.display(this);
			});
			this.main_window.troubleshooting_action.connect(() => {
				TroubleshootingWindow.display(this);
			});
			var sw = new ScrolledWindow(null, null);
			sw.set_policy (PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
			sw.add(this.frame);
			sw.show_all();
			
			this.main_window.add(sw);
			
			/* Show page title as window title */
			this.frame.title_changed.connect((source, frame, title) => {
				this.main_window.title = title;
			});
		}
		
		/**
		 * Launch application interface
		 */
		public override void run(){
			this.frame.start();
			this.player.start();
#if GLIB_DBUS
			this.mpris_manager.start();
			this.dock_manager.start();
#endif
			// TrayIcon must not be shown before MPRIS interface is initialized!
			Idle.add(() => {
				this.tray_icon.show();
				return false;
			});
#if UNITY_QUICK_LIST
	this.unity_quick_list = new GMF.Helpers.UnityLauncher(this, this.player);
	unity_quick_list.start();
#endif			
				
			/* Handle keybindings */
			media_keys = new GMF.MediaKeys.Manager(this.app_name, this.player);
			bool pause_key = this.config.bool("pause_key", false);
			bool multimedia_keys = this.config.bool("multimedia_keys", true);
			if(pause_key || multimedia_keys){
				if(pause_key) media_keys.handle_pause_key();
				if(multimedia_keys) media_keys.handle_multimedia_keys();
			}
			
			this.main_window.show_all();
			this.frame.grab_focus();
			
			var prev_release = this.ui_state["release"] ?? "0.1";
			if(prev_release != GMF.Config.RELEASE){
				ui_state["release"] = GMF.Config.RELEASE;
				try{
					ui_state.save();
				}
				catch(Error e){
					log_critical("Unable to save UI state: %s", e.message);
				}
				this.main_window.show_about_dialog();
			}
#if DEPRECATED_PPA
			if(ui_state["disable_ppa_warning_" + GMF.Config.DEPRECATED_PPA] == null){
				string new_ppa;
				string new_ppa_link;
				string will_not;
				switch(GMF.Config.DEPRECATED_PPA){
				case "janousek.jiri/google-music-frame-daily":
					new_ppa = "nuvola-player-builders/unstable";
					new_ppa_link = "https://launchpad.net/~nuvola-player-builders/+archive/unstable";
					will_not = "updated anymore";
					break;
				case "janousek.jiri/google-music-frame-releases":
					new_ppa = "nuvola-player-builders/stable";
					new_ppa_link = "https://launchpad.net/~nuvola-player-builders/+archive/stable";
					will_not = "upgraded to the release 0.4";
					break;
				default:
					assert_not_reached();
				}
				var dialog = new Gtk.MessageDialog.with_markup(this.main_window,
					Gtk.DialogFlags.MODAL, Gtk.MessageType.OTHER, Gtk.ButtonsType.OK,
			"""<b>You are running a build from a deprecated PPA ppa:%s</b>

Please switch to the new PPA: <a href="%s">ppa:%s</a>

The deprecated PPA will not be %s and will be eventually removed.""".printf(
					GMF.Config.DEPRECATED_PPA, new_ppa_link, new_ppa, will_not));
				var checkbox = new Gtk.CheckButton.with_label("Do not show this warning again");
				checkbox.show();
				checkbox.toggled.connect( () => {
					ui_state["disable_ppa_warning_" + GMF.Config.DEPRECATED_PPA] =
					checkbox.active ? "true" : null; 
					try{
						ui_state.save();
					}
					catch(Error e){
						log_warning("Unable to sace UI state: %s", e.message);
					}
				});
				dialog.action_area.pack_start(checkbox, false, false, 0);
				dialog.action_area.reorder_child(checkbox, 0);
				var img_file = Fenryxo.Path.get_data_file(this.app_name + "/images/gmf_renamed_to_nuvola.png");
				if(img_file != null || true){
					var box = dialog.get_content_area() as Gtk.Box;
					var img = new Gtk.Image.from_file(img_file.get_path());
					img.show();
//~ 					dialog.image = img;
					box.pack_start(img, false, false, 5);
					box.reorder_child(img, 0);
				}
				dialog.run();
				dialog.hide();
				dialog.destroy();
			}
#endif
		}
		
		/**
		 * Emitted when application is about to be activated 
		 * (e.g. hidden window will be presented)
		 */
		public override void activated(){
			log_debug("Application activated");
			this.main_window.present();
		}
		
		/**
		 * Emitted when application is about to be terminated
		 */
		public override void terminated(){
			log_debug("Application terminated");
			Gtk.main_quit();
		}
		
		private void on_close_action(){
			if(!this.can_hide){
				log_debug("On close: quit");
				try{
					this.quit();
				}
				catch(Fenryxo.AppError e){
					log_error("Unable to quit: %s", e.message);
				}
			}
			else{
				log_debug("On close: hide");
			}
		}
		
		private void config_changed(){
			this.hide_when_playing = this.config.bool("hide_when_playing", true);
			bool pause_key = this.config.bool("pause_key", false);
			bool multimedia_keys = this.config.bool("multimedia_keys", true);
			if(pause_key)
				this.media_keys.handle_pause_key();
			else
				this.media_keys.release_pause_key();
			if(multimedia_keys)
				this.media_keys.handle_multimedia_keys();
			else
				this.media_keys.release_multimedia_keys();
			
		}
		
		/**
		 * Purge cookies and reload the web interface.
		 */
		public void purge_cookies(){
			set_cookies(this.storage, true);
			this.frame.reload();
		}
		
		/**
		 * Purge album art cache and reload the web interface.
		 */
		public void purge_album_art(){
			this.storage.purge_album_art();
			this.frame.reload();
		}
		
#if LIBSOUP_CACHE
		/**
		 * Purge browser cache reload the web interface.
		 */
		public void purge_browser(){
			set_cache(this.storage, true);
			this.frame.reload_bypass_cache();
		}
#endif
	}
}
