/*
Google Music Frame :: Main entry point

This file contains main entry point of application. It's responsible for
  * parsing command line arguments
  * environment settings (log level, configuration)
  * preparing Soup session (cookies storage, proxy settings)
  * launching main application class
   

Copyright 2011 Jiří Janoušek <janousek.jiri@gmail.com>

This program is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License version 3, as published 
by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranties of 
MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using Fenryxo;
using Fenryxo.Logger;
using Fenryxo.Config;
using Fenryxo.Path;
using Soup;
using GMF;

[CCode (cheader_filename = "sys/prctl.h", cname = "prctl")]
extern int prctl (int option, string arg2, ulong arg3, ulong arg4, ulong arg5);

/**
 * Struct containing command line arguments. Check Args.options for their meaning.
 */
struct Args {
	static bool debug;
	static bool verbose;
	static bool version;
	static bool quit;
	
	public static const OptionEntry[] options = {
		{ "version", 'v', 0, GLib.OptionArg.NONE, ref Args.version, "Show version of program", null },
		{ "verbose", 'V', 0, GLib.OptionArg.NONE, ref Args.verbose, "Print information to terminal", null },
		{ "debug", 'D', 0, GLib.OptionArg.NONE, ref Args.debug, "Print debug messages", null },
		{ "quit", 'Q', 0, GLib.OptionArg.NONE, ref Args.quit, "Quit currently running instance", null },
		{ null }
	};
}

/* Setup internet connection */
void set_proxy(ConfigDict config){
	var session = WebKit.get_default_session();
	var use_proxy = config["use_proxy"];
	string? proxy = null;
	if(use_proxy == "manual"){
		proxy = config["proxy"];
		if(proxy != null && proxy != "")
			log_message("Setting proxy (manual): %s", proxy);
		else
			log_critical("Unable to apply manual proxy: proxy option is empty.");
	}
	else if(use_proxy == null || use_proxy == "" || use_proxy == "auto"){
		proxy = Environment.get_variable("https_proxy")
		        ?? Environment.get_variable("http_proxy");
		if(proxy != null && proxy != "")
			log_message("Setting proxy (auto): %s", proxy);
		else
			log_debug("No environent proxy settings.");
	}
	else
		log_debug("No proxy settings applied.");
	
	session.abort();
	if(proxy == null || proxy == ""){
		session.proxy_uri = null;
	}
	else{
		session.proxy_uri = new Soup.URI(proxy);
	}
}

/* Setup Cookies storage */
void set_cookies(Storage storage, bool clear=false){
	var file = storage.get_default_config_file(GMF.Config.COOKIES_FILE);
	if(clear){
		try{
			file.delete();
		}
		catch(Error e){
			log_warning("Unable to delete cookies file: %s", file.get_path());
		}
	}
	weak Session session = WebKit.get_default_session();
	weak SessionFeature? feature = session.get_feature(typeof(CookieJarText));
	if(feature != null) session.remove_feature(feature);
	session.add_feature = new CookieJarText(file.get_path(), false);
}

#if LIBSOUP_CACHE
/* Setup browser cache */
void set_cache(Storage storage, bool clear=false){
	var dir = storage.get_default_cache_dir("browser");
	weak Session session = WebKit.get_default_session();
	SessionFeature? feature = session.get_feature(typeof(Cache));
	if(feature == null){
		feature = new Cache(dir.get_path());
		(feature as Cache).load();
		session.add_feature = (owned) feature;
	}
	if(clear){
		weak Cache cache = feature as Cache;
		cache.clear();
		cache.dump();
		cache.flush();
	}
}

void save_cache(){
	log_debug("Dump cache");
	weak Session session = WebKit.get_default_session();
	SessionFeature? feature = session.get_feature(typeof(Cache));
	if(feature == null) return;
	weak Cache cache = feature as Cache;
	cache.dump();
	cache.flush();
}
#endif

/**
 * Main entry point
 */
int main(string[] args){
	var command = string.joinv(" ", args);
	
	/* Parse command line arguments */
	var opt_context = new OptionContext("");
	try{
		opt_context.set_help_enabled(true);
		opt_context.add_main_entries(Args.options, null);
		opt_context.parse(ref args);
    }
    catch(OptionError e) {
        stderr.printf("Error: Option parsing failed: %s\n\n%s\n",
			e.message, opt_context.get_help(true, null));
        return 1;
    }
    
    // Show version and exit
    if(Args.version){
		stdout.printf("%s %s\n", GMF.Config.DISPLAY_NAME, GMF.Config.VERSION);
		return 0;
	}
	
	
	
	/* Logger initialization.
	 * Use Fenryxo.log_debug() instead of GLib.debug() etc. to use "GMF" domain.
	 */
	Fenryxo.Logger.init("GMF", Args.debug ? LogLevelFlags.LEVEL_DEBUG
		: (Args.verbose ? LogLevelFlags.LEVEL_INFO : LogLevelFlags.LEVEL_WARNING));
	
	/* 
	 * If we are running from source directory, append source data directory to
	 * XDG_DATA_DIRS
	 */
	File? dir = File.new_for_path("./");
	if(dir.get_basename() == "build") dir = dir.get_parent();
	else if(!dir.get_child("build").query_exists()) dir = null;
	
	if(dir != null){
		string? dirs = Environment.get_variable("XDG_DATA_DIRS");
		string data_dir = dir.get_child("data").get_path();
		log_debug("Appending %s to XDG_DATA_DIRS", data_dir);
		if(dirs == null){
			Environment.set_variable("XDG_DATA_DIRS", data_dir, true);
		}
		else{
			Environment.set_variable("XDG_DATA_DIRS", data_dir + ":" + dirs, true);
		}
		
	}
	
	log_debug("Gettext: %s, %s", GMF.Config.APPNAME, GMF.Config.LOCALES_DIR);
	log_debug("Gettext metadata: %s", _(""));
	Intl.textdomain(GMF.Config.APPNAME);
	Intl.bindtextdomain(GMF.Config.APPNAME, GMF.Config.LOCALES_DIR);
	
	Gtk.init(ref args);
	prctl(15, GMF.Config.BINARY_NAME, 0, 0, 0);
	GLib.Environment.set_application_name(GMF.Config.BINARY_NAME); // doesn't work?
	Gdk.set_program_class(GMF.Config.DISPLAY_NAME); // must be set after Gtk.init()!
	
	
	var storage = new GMF.Storage(GMF.Config.APPNAME);
	var config = new ConfigFile(storage.get_default_config_file(GMF.Config.CONFIG_FILE));
	var ui_state = new ConfigFile(storage.get_default_config_file(GMF.Config.UI_STATE_FILE));
	

	
	var app = new GMF.Application(config, ui_state, storage);
	if(Args.quit){
		// Try to quit the running instance
		try{
			app.quit();
			return 0;
		}
		catch(AppError e){
			if(e is AppError.NOT_RUNNING){
				stderr.printf("No instance is running.\n");
				return 1;
			}
			if(e is AppError.UNABLE_TO_TERMINATE){
				stderr.printf("Unable to terminate running instance.\n");
				return 1;
			}
			stderr.printf("Unknown error: %s\n", e.message);
			return 1;
		}
	}
	
	
	// Try to activate the running instance
	try{
		if(app.activate()){
			// There is a running instance and it has been activated => exit
			return 0;
		}
	}
	catch(AppError e){
		if(e is AppError.UNABLE_TO_ACTIVATE){
			stderr.printf("Unable to activate running instance.\n");
			return 1;
		}
		stderr.printf("Unknown error: %s\n", e.message);
		return 1;
	}
		
	// No instance is already running => continue
	log_message("Welcome to %s, version: %s", GMF.Config.DISPLAY_NAME, GMF.Config.VERSION);
#if DEV_BUILD
	log_message("You are running development version");
#endif
	log_message("Report any issues/bugs you might find to %s", GMF.Config.BUG_TRACKER);
	log_debug("command: %s", command);
#if GLIB_DBUS
	log_debug("Built with GLib DBus");
#endif
#if LIBSOUP_CACHE
	log_debug("Built with libsoup cache");
#endif
#if INDICATOR_SOUND_MAVERICK
	log_debug("Built with Indicator sound Maverick");
#endif
#if UNITY_QUICK_LIST
	log_debug("Built with Unity Quick List");
#endif

	
	
	
	
	/* GMF remembers cookies to store session */
	set_cookies(storage);
#if LIBSOUP_CACHE
	set_cache(storage);
	app.terminated.connect(save_cache);
#endif
	set_proxy(config);
	// Runtime proxy settings is affected by a bug in libsoup, may be fixed in libsoup 2.34.1
	// http://osdir.com/ml/commits.gnome/2011-04/msg09714.html
	// http://ftp.acc.umu.se/pub/gnome/sources/libsoup/2.34/libsoup-2.34.1.news
	var libsoup_version = GMF.Config.LIBSOUP_2_4_VERSION.split(".");
	if(int.parse(libsoup_version[0]) >= 2
	&& int.parse(libsoup_version[1]) >= 34
	&& int.parse(libsoup_version[2]) >= 1){
		log_debug("Libsoup >= 2.34.1, runtime proxy settings can be enabled.");
		// TODO: check if the bug is really fixed
		#if DEV_BUILD
		config.notify_changed.connect(set_proxy);
		#endif
	}
	
	/* Let's launch main class */
	try{
		if(!app.start()){
			// There is already a running instance => exit
			return 0;
		}
	}
	catch(AppError e){
		if(e is AppError.UNABLE_TO_ACTIVATE){
			stderr.printf("Unable to activate running instance.\n");
			return 1;
		}
		stderr.printf("Unknown error: %s\n", e.message);
		return 1;
	}


	// Changes since last release
//~ 	GMF.show_changelog(ui_state);
	Gtk.main();
	return 0;
}
