/*
Google Music Frame :: Application

This file contains main application class and its wrapper for DBus.

Copyright 2011 Jiří Janoušek <janousek.jiri@gmail.com>

This program is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License version 3, as published 
by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranties of 
MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using Fenryxo.Logger;
using Fenryxo;
using GMF;

namespace GMF.MediaKeys{
	
	/**
	 * Manages multimedia keys and pause key
	 */
	public class Manager : GLib.Object{
		private KeybindingManager keybindings;
#if GLIB_DBUS
		private GnomeMedia media_keys;
#endif
		private string app_name;
		private GMF.Player player;
		
		/**
		 * Construct multimedia keys manager
		 * 
		 * @param app_name Application name
		 * @param player Player controling playback
		 */
		public Manager(string app_name, GMF.Player? player){
			this.app_name = app_name;
			this.player = player;
		}
		
		/**
		 * Binds Pause key to play/pause action
		 */
		public void handle_pause_key(){
			if(this.keybindings == null) this.keybindings = new KeybindingManager();
			this.keybindings.bind("Pause", () => {this.player.toggle_play();});
		}
		
		/**
		 * Binds multimedia keys to control playback
		 */
		public void handle_multimedia_keys(){
#if GLIB_DBUS
			/* Watch sound menu */
			Bus.watch_name(BusType.SESSION, "org.gnome.SettingsDaemon",
				BusNameWatcherFlags.NONE,
				this.gnome_settings_appeared, this.gnome_settings_vanished);
#else
			this.media_keys_fallback();
#endif
		}
		
		
		/**
		 * Unbinds multimedia keys to stop controlling playback
		 */
		public void release_multimedia_keys(){
#if GLIB_DBUS
			this.media_keys_stop_fallback();
			if(this.media_keys == null) return;
			try{
				this.media_keys.release_media_player_keys(this.app_name);
				this.media_keys.media_player_key_pressed.disconnect(this.on_media_key_pressed);
				this.media_keys = null;
				
			}
			catch(IOError e){
				log_warning("Unable to get proxy for GNOME Media keys: %s", e.message);
				this.media_keys = null;
			}
#else
			this.media_keys_stop_fallback();
#endif
		}
		
		/**
		 * Unbinds Pause key from play/pause action
		 */
		public void release_pause_key(){
			if(this.keybindings != null) this.keybindings.unbind("Pause");
		}
		
		
		
#if GLIB_DBUS
		/**
		 * Use GNOME settings daemon to control multimedia keys
		 */
		private void gnome_settings_appeared(DBusConnection conn, string name, string owner){
			log_debug("GNOME settings daemon appeared: %s, %s", name, owner);
			this.media_keys_stop_fallback();
			try{
				this.media_keys = Bus.get_proxy_sync(BusType.SESSION,
				"org.gnome.SettingsDaemon",
				"/org/gnome/SettingsDaemon/MediaKeys");
				this.media_keys.grab_media_player_keys(this.app_name, 0);
				this.media_keys.media_player_key_pressed.connect(this.on_media_key_pressed);
				
			}
			catch(IOError e){
				log_warning("Unable to get proxy for GNOME Media keys: %s", e.message);
				this.media_keys = null;
			}
		}
		
		private void gnome_settings_vanished(DBusConnection conn, string name){
			log_debug("GNOME settings daemon vanished: %s", name);
			this.media_keys = null;
			this.media_keys_fallback();
			
			
		}
		
		/**
		 * Triggers media keys actions
		 */
		private void on_media_key_pressed(string app_name, string key){
			log_debug("Media key pressed: %s, %s", app_name, key);
			if(app_name != this.app_name) return;
			switch(key){
				case "Play":
				case "Pause": this.player.toggle_play(); break;
				case "Stop": this.player.stop(); break;
				case "Previous": this.player.previous_song(); break;
				case "Next": this.player.next_song(); break;
			}
		}
#endif		
		/**
		 * Fallback to use Xorg keybindings
		 */
		private void media_keys_fallback(){
			if(this.keybindings == null) this.keybindings = new KeybindingManager();
			this.keybindings.bind("XF86AudioPlay", () => {this.player.toggle_play();});
			this.keybindings.bind("XF86AudioPause", () => {this.player.toggle_play();});
			this.keybindings.bind("XF86AudioStop", () => {this.player.stop();});
			this.keybindings.bind("XF86AudioPrev", () => {this.player.previous_song();});
			this.keybindings.bind("XF86AudioNext", () => {this.player.next_song();});
		}
		
		private void media_keys_stop_fallback(){
			if(this.keybindings == null) return;
			this.keybindings.unbind("XF86AudioPlay");
			this.keybindings.unbind("XF86AudioPause");
			this.keybindings.unbind("XF86AudioStop");
			this.keybindings.unbind("XF86AudioPrev");
			this.keybindings.unbind("XF86AudioNext");
		}
	}

#if GLIB_DBUS	
	[DBus(name = "org.gnome.SettingsDaemon.MediaKeys")]
	public interface GnomeMedia: Object{
		public abstract void grab_media_player_keys(string app, uint32 time) throws IOError;
		public abstract void release_media_player_keys(string app) throws IOError;
		public signal void media_player_key_pressed(string app, string key);
	}
#endif	
}
