/*
Google Music Frame

Copyright 2011 Jiří Janoušek <janousek.jiri@gmail.com>

This program is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License version 3, as published 
by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranties of 
MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using Fenryxo.Logger;


/**
 * Player stores and controls playback state. It's also responsible for notifications
 */
public class GMF.Player : GLib.Object{
	private GMF.Application app;
	private Storage storage;
	private Frame frame;
	private bool skip_notification = false;
	private Notify.Notification notification;
	
	[Description(nick = "availability of action 'go to previous song'", blurb = "availability of action 'go to previous song'")]
	public bool can_prev {get; private set; default = false;}
		
	[Description(nick = "availability of action 'go to next song'", blurb = "availability of action 'go to next song'")]
	public bool can_next {get; private set; default = false;} 
	
	[Description(nick = "playback state", blurb = "playback state: 'playing', 'paused' or 'none'")]
	public string? playback_state {get; private set; default = "none";} 
	
	[Description(nick = "URL of album art", blurb = "URL of album art")]
	private string? album_art_url {get; private set;} 
	
	[Description(nick = "Path to album art", blurb = "Path to album art")]
	private string? album_art {get; private set;} 
	
	[Description(nick = "name of the current artist", blurb = "name of the current artist")]
	public string? artist {get; private set;} 
	
	[Description(nick = "name of the current song", blurb = "name of the current song")]
	public string? song {get; private set;} 
	
	
	/**
	 * Constructs new player controling given frame.
	 * 
	 * @param app Application object
	 * @param frame Frame with Google Music web interface
	 * @param storage Storage for album art
	 */
	public Player(GMF.Application app, Frame frame, Storage storage){
		this.app = app;
		this.frame = frame;
		this.storage = storage;
	}
	
	/**
	 * Shows initial notification with program name and version and starts to response
	 * to frame events.
	 */
	public void start(){
		Notify.init(this.app.display_name);
		this.app.terminated.connect(() => {Notify.uninit();});
		this.notification = new Notify.Notification(this.app.display_name,
		/// Initial notification
		                    _("Version") + " " + this.app.version, this.app.icon);
		this.frame.data_changed.connect(this.on_frame_state_changed);
		try{
			this.notification.show();
		}
		catch(Error e){
			log_warning("Unable to show initial notification");
		}
	}
	
	/**
	 * Emitted when song is changed.
	 * 
	 * @param art album art file
	 * @param artist name of the current artist
	 * @param song name of the current song 
	 */
	public virtual signal void song_changed(string album_art, string? artist, string? song){
		if(song == null || song == "") return;
		/// Notifications
		this.notification.update(song,
			(artist == null || artist == "")
			/// Notifications
			? _("unknown artist")
			/// Notifications - %s will be replaced with artist
			: _("by %s").printf(artist), album_art);
	}
	
	/**
	 * Emitted when playback state is changed.
	 * 
	 * @param state playback state: "playing", "paused" or "none"
	 * @param can_prev availability of action "go to previous song" 
	 * @param can_next availability of action "go to next song"
	 */
	public signal void playback_state_changed(string state, bool can_prev, bool can_next);
	
	
	/**
	 * Response to frame events - updates current status, stores album art
	 * and emit associated signals
	 */
	private void on_frame_state_changed(string playback_state, bool can_prev,
	bool can_next, string? art_uri, string? artist, string? song){
		var state_changed = false;
		var song_changed = false;
		if(this.playback_state != playback_state){
			this.playback_state = playback_state;
			state_changed = true;
		}
		
		if(this.can_prev != can_prev){
			this.can_prev = can_prev;
			state_changed = true;
		}
		
		if(this.can_next != can_next){
			this.can_next = can_next;
			state_changed = true;
		}
		
		if(this.album_art_url != art_uri){
			this.album_art_url = art_uri;
			this.album_art = this.storage.get_album_art(art_uri) ?? this.app.icon;
			song_changed = true;
		}
		
		if(this.artist != artist){
			this.artist = artist;
			song_changed = true;
		}
		
		if(this.song != song){
			this.song = song;
			song_changed = true;
		}
		
		if(song_changed){
			this.song_changed(album_art, artist, song);
		}
		
		if(state_changed){
			this.playback_state_changed(playback_state, can_prev, can_next);
		}
		
		if(song_changed || state_changed){
			this.show_notification();
		}
	}
	
	/**
	 * Skips to the next track in the tracklist.
	 * 
	 * If there is no next track (and endless playback and track repeat are both off),
	 * playback will be stopped. If playback is paused or stopped, it remains that way.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public virtual signal void next_song(bool notify=true){
		try{
			this.skip_notification = ! notify;
			this.frame.next_song();
		}
		catch(FrameError e){
			log_critical("Player: nextSong command failed: %s", e.message);
			if(!notify) this.skip_notification = false;
		}
	}
	
	/***
	 * Skips to the previous track in the tracklist.
	 * 
	 * If there is no previous track (and endless playback and track repeat are both off),
	 * playback will be stopped. If playback is paused or stopped, it remains that way.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void previous_song(bool notify=true){
		try{
			this.skip_notification = ! notify;
			this.frame.prev_song();
		}
		catch(FrameError e){
			log_critical("Player: prevSong command failed: %s", e.message);
			if(!notify) this.skip_notification = false;
		}
	}
	
	/**
	 * Pauses playback.
	 * 
	 * If playback is already paused, this has no effect. Calling Play after this should
	 * cause playback to start again from the same position.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void pause(bool notify=true){
		/* No notifications are shown on pause action,
		 * therefore notify param has no effect
		 */
		try{
			this.frame.pause();
		}
		catch(FrameError e){
			log_critical("Player: pauseSong command failed: %s", e.message);
		}
	}
	
	/**
	 * Pauses playback.
	 * If playback is already paused, resumes playback.
	 * If playback is stopped, starts playback.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void toggle_play(bool notify=true){
		try{
			if(this.playback_state != "playing") this.skip_notification = ! notify;
			this.frame.toggle_play();
		}
		catch(FrameError e){
			log_critical("Player: playPause command failed: %s", e.message);
			if(!notify) this.skip_notification = false;
		}
	}
	
	/**
	 * Stops playback.
	 * 
	 * If playback is already stopped, this has no effect. Calling Play after this should
	 * cause playback to start again from the beginning of the track. 
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void stop(bool notify=true){
		/* No notifications are shown on pause action,
		 * therefore notify param has no effect
		 */
		try{
			this.frame.stop();
		}
		catch(FrameError e){
			log_critical("Player: pauseSong command failed: %s", e.message);
		}
	}
	
	/**
	 * Starts or resumes playback.
	 * 
	 * If already playing, this has no effect.
	 * If there is no track to play, this has no effect.
	 * 
	 * @param notify if true a notification will be shown if needed
	 */
	public void play(bool notify=true){
		try{
			this.skip_notification = ! notify;
			this.frame.play();
		}
		catch(FrameError e){
			log_critical("Player: playPause command failed: %s", e.message);
			if(!notify) this.skip_notification = false;
		}
	}
	
	public void thumbs_up(bool notify=true){
		try{
//~ 			this.skip_notification = ! notify;
			this.frame.thumbs_up();
		}
		catch(FrameError e){
			log_critical("Player: thumbUp command failed: %s", e.message);
//~ 			if(!notify) this.skip_notification = false;
		}
	}
	
	public void thumbs_down(bool notify=true){
		try{
//~ 			this.skip_notification = ! notify;
			this.frame.thumbs_down();
		}
		catch(FrameError e){
			log_critical("Player: thumbUp command failed: %s", e.message);
//~ 			if(!notify) this.skip_notification = false;
		}
	}
	
	/**
	 * Shows notification if needed
	 */
	private void show_notification(){
		if(this.playback_state != "playing") return;
		
		if(!this.skip_notification){
			try{
				this.notification.show();
			}
			catch(Error e){
				log_warning("Unable to send notification: %s", e.message);
			}
		}
		else if(this.skip_notification){
			this.skip_notification = false;
		}	
	}
}
